import unittest
import random
from itertools import product
from collections import defaultdict
from sklearn.datasets import make_classification

import numpy as np
from scipy.stats import spearmanr
import xgboost as xgb
from src.vertibench.Splitter import ImportanceSplitter, CorrelationSplitter, SimpleSplitter
from src.vertibench.Evaluator import ImportanceEvaluator, CorrelationEvaluator


def generate_data():
    X_1_10, y_1_10 = make_classification(n_samples=1, n_features=10, n_informative=10, n_redundant=0)
    X_1_500, y_1_500 = make_classification(n_samples=1, n_features=500, n_informative=20, n_redundant=0)
    X_10_10, y_10_10 = make_classification(n_samples=10, n_features=10, n_informative=10, n_redundant=0)
    X_1k_10, y_1k_10 = make_classification(n_samples=1000, n_features=10, n_informative=10, n_redundant=0)
    X_1k_500, y_1k_500 = make_classification(n_samples=1000, n_features=500, n_informative=20, n_redundant=0)

    data_dict = {
        "1_10": (X_1_10, y_1_10),
        "1_500": (X_1_500, y_1_500),
        "10_10": (X_10_10, y_10_10),
        "1k_10": (X_1k_10, y_1k_10),
        "1k_500": (X_1k_500, y_1k_500),
    }
    return data_dict


class TestImportanceSplitter(unittest.TestCase):
    def setUp(self):
        self.n_parties_list = [0, 1, 2, 4, 32, 512]
        self.const_weight_list = [0.0, 1e-6, 1e-3, 1, 1e3, 1e6]
        np.random.seed(0)
        self.weights10_list = np.random.exponential(5, size=(5, 10))    # 5 different weights for 10 parties
        self.weights1k_list = np.random.exponential(5, size=(5, 1000))  # 5 different weights for 1000 parties
        self.rounds = 1000
        self.data_dict = generate_data()

    def test_dirichlet(self):
        # the ImportanceSplitter.dirichlet function should has the same distribution as np.random.dirichlet

        for sz in range(2, 100, 20):
            Xs = []
            Ys = []
            for _ in range(10000):
                alpha = np.random.exponential(5, size=sz).flatten()
                alpha = alpha / np.sum(alpha)
                xs = np.random.dirichlet(alpha)
                if np.isnan(xs).any():
                    continue
                ys = ImportanceSplitter.dirichlet(alpha)

                Xs.append(xs)
                Ys.append(ys)

            Xs = np.array(Xs)
            Ys = np.array(Ys)
            bins = np.arange(0, 1.01, 0.05)
            for col in range(sz):
                hist_Xs = np.histogram(Xs[:, col], bins=bins)[0] / Xs.shape[0]
                hist_Ys = np.histogram(Ys[:, col], bins=bins)[0] / Ys.shape[0]
                self.assertTrue(np.allclose(hist_Xs, hist_Ys, atol=2e-2), msg=f"col={col}, sz={sz}, "
                                                                              f"diff={abs(hist_Xs - hist_Ys)}")

    def check_params(self):
        splitters = []
        for n_parties, const_weight in product(self.n_parties_list, self.const_weight_list):
            if n_parties <= 1 or const_weight <= 0.0:
                with self.assertRaises(ValueError):
                    ImportanceSplitter(n_parties, const_weight)
            else:
                splitters.append(ImportanceSplitter(n_parties, const_weight))

        n_parties = 10
        for weights in self.weights10_list:
            if any([weight <= 0.0 for weight in weights]):
                with self.assertRaises(ValueError):
                    ImportanceSplitter(n_parties, weights=weights)
            splitters.append(ImportanceSplitter(n_parties, weights=weights))

        return splitters

    def test_split_indices(self):
        splitters = self.check_params()
        for key, (X, y) in self.data_dict.items():
            # train a model on each split
            model = xgb.XGBClassifier()
            model.fit(X, y)
            feature_importances = np.array(model.feature_importances_)

            const_weight_stds_all_parties = defaultdict(list)
            for splitter in splitters:
                if splitter.num_parties > X.shape[1]:
                    continue
                with self.subTest(key=key, n_parties=splitter.num_parties):
                    std_feature_cnt_summary = []
                    std_summary = []
                    mean_by_party_summary = [0 for _ in range(splitter.num_parties)]
                    for _ in range(self.rounds):
                        split_indices = splitter.split_indices(X, allow_empty_party=False)
                        self.assertEqual(len(split_indices), splitter.num_parties)
                        self.assertEqual(sum([len(indices) for indices in split_indices]), X.shape[1])
                        for indices in split_indices:
                            self.assertGreater(len(indices), 0) # no empty party
                        importance_per_party = []
                        for party_id, indices in enumerate(split_indices):
                            score = np.sum(feature_importances[indices])
                            importance_per_party.append(score)
                            mean_by_party_summary[party_id] += score
                        std_summary.append(np.std(importance_per_party, ddof=1))
                        std_feature_cnt_summary.append(np.std([len(indices) for indices in split_indices], ddof=1))
                    mean_by_party_summary = np.array([score / self.rounds for score in mean_by_party_summary])

                    if X.shape[0] <= 1:
                        continue

                    # mean_by_party_summary should be proportional to weights
                    self.assertEqual(len(mean_by_party_summary), len(splitter.weights))
                    self.assertGreater(np.sum(mean_by_party_summary), 0.0, msg=f"key={key}, weights={splitter.weights},"
                                                                               f" mean_by_party_summary={mean_by_party_summary}")
                    scaled_importance = mean_by_party_summary / np.sum(mean_by_party_summary)
                    scaled_weights = splitter.weights / np.sum(splitter.weights)
                    rmse = np.sqrt(np.mean((scaled_importance - scaled_weights) ** 2))
                    self.assertLessEqual(rmse, 0.1, msg=f"{scaled_importance=}, {scaled_weights=}")

                    if np.all(np.isclose(splitter.weights, splitter.weights[0])):
                        const_weight_stds_all_parties[splitter.num_parties].append(
                            (splitter.weights[0], np.mean(std_feature_cnt_summary), np.mean(std_summary)))

            for n_parties, const_weight_stds in const_weight_stds_all_parties.items():
                with self.subTest(key=key, n_parties=n_parties):
                    if len(const_weight_stds) == 0:
                        continue
                    const_weight_stds = np.array(const_weight_stds)
                    # three columns should be ordered
                    order0 = np.argsort(const_weight_stds[:, 0])[::-1]
                    order1 = np.argsort(const_weight_stds[:, 1])
                    order2 = np.argsort(const_weight_stds[:, 2])
                    cor1 = spearmanr(order0, order1)[0]
                    cor2 = spearmanr(order0, order2)[0]
                    self.assertGreater(cor1, 0.6, msg=f"key={key}, cor1={cor1}, cor2={cor2}")
                    self.assertGreater(cor2, 0.6, msg=f"key={key}, cor1={cor1}, cor2={cor2}")

    def test_split_tabular(self):
        # assuming that the split_indices function is correct

        # test with/without given indices
        for key, (X, y) in self.data_dict.items():
            if X.shape[1] < 10:
                continue
            with self.subTest(task="tabular", key=key):
                splitter1 = ImportanceSplitter(10, seed=0)
                split_indices = splitter1.split_indices(X, allow_empty_party=False)
                Xs1 = splitter1.split(X, indices=split_indices)
                splitter2 = ImportanceSplitter(10, seed=0)
                Xs2 = splitter2.split(X)
                self.assertEqual(len(Xs1), 10)
                self.assertEqual(len(Xs2), 10)
                for i in range(10):
                    self.assertTrue(np.allclose(Xs1[i], Xs2[i]))

    def test_split_tabular_fill(self):
        # assuming that the split_indices function is correct

        # test with/without given indices
        for key, (X, y) in self.data_dict.items():
            if X.shape[1] < 10:
                continue
            with self.subTest(task="fill", key=key):
                splitter1 = ImportanceSplitter(10, seed=0)
                split_indices = splitter1.split_indices(X)
                Xs1 = splitter1.split(X, indices=split_indices, fill=-1)
                splitter2 = ImportanceSplitter(10, seed=0)
                Xs2 = splitter2.split(X, fill=-1)
                self.assertEqual(len(Xs1), 10)
                self.assertEqual(len(Xs2), 10)
                for i in range(10):
                    self.assertTrue(np.allclose(Xs1[i], Xs2[i]))

                # check if the fill value is correct
                for i in range(X.shape[0]):
                    for X1 in Xs1:
                        fill_cnt = np.count_nonzero(X1[i] == -1)
                        self.assertLessEqual(fill_cnt, X.shape[1] * (10 - 1))

    def test_split_multi_tabular(self):
        # assuming that the split_indices function is correct

        # test with/without given indices
        X_list = []
        for key, (X, y) in self.data_dict.items():
            if X.shape[1] != 10:
                continue
            X_list.append(X)

        with self.subTest(task="multi_tabular"):
            splitter = ImportanceSplitter(10, seed=0)
            Xs = splitter.split(*X_list)
            self.assertEqual(len(Xs), len(X_list))
            for i in range(len(X_list)):
                self.assertEqual(len(Xs[i]), 10)

            # concat all the parties should be the same as the original data
            X_all = np.concatenate(X_list, axis=0)
            Xs_all = []
            for i in range(len(X_list)):
                Xs_all.append(np.concatenate([Xs[i][j] for j in range(10)], axis=1))
            X_all_split = np.concatenate(Xs_all, axis=0)
            self.assertTrue(np.allclose(np.sort(X_all), np.sort(X_all_split)))

    def test_split_multi_tabular_fill(self):
        # assuming that the split_indices function is correct

        # test with/without given indices
        X_list = []
        for key, (X, y) in self.data_dict.items():
            if X.shape[1] != 10:
                continue
            X_list.append(X)

        with self.subTest(task="multi_tabular_fill"):
            splitter = ImportanceSplitter(10, seed=0)
            Xs = splitter.split(*X_list, fill=-1)
            self.assertEqual(len(Xs), len(X_list))
            for i in range(len(X_list)):
                self.assertEqual(len(Xs[i]), 10)

            # concat all the parties should be the same as the original data
            X_all = np.concatenate(X_list, axis=0)
            Xs_all = []
            for i in range(len(X_list)):
                Xs_all.append(np.concatenate([Xs[i][j] for j in range(10)], axis=1))
            X_all_split = np.concatenate(Xs_all, axis=0)
            self.assertEqual(X_all_split.shape[1], X_all.shape[1] * 10)
            X_all_split_filled = X_all_split[X_all_split != -1].reshape(X_all.shape)
            self.assertTrue(np.allclose(np.sort(X_all), np.sort(X_all_split_filled)))


class TestCorrelationSplitter(unittest.TestCase):
    def setUp(self):
        self.n_parties_list = [0, 1, 2, 4, 32, 512]
        self.beta_list = [0.0, 1e-6, 0.5, 1, 1.1]

        self.rounds = 1000
        self.data_dict = generate_data()

    def test_fit_split(self):
        for key, (X, y) in self.data_dict.items():
            if X.shape[1] <= 1 or X.shape[0] <= 2:
                continue
            for n_parties in self.n_parties_list:
                for beta in self.beta_list:
                    with self.subTest(key=key, n_parties=n_parties, beta=beta):
                        splitter = CorrelationSplitter(n_parties, gpu_id=0)

                        if beta > 1.0 or beta < 0.0:
                            with self.assertRaises(ValueError):
                                splitter.fit_split(X, beta=beta)
                        Xs = splitter.fit_split(X, beta=beta)
                        beta_eval = splitter.evaluator.evaluate_beta(Xs)
                        self.assertAlmostEqual(beta, beta_eval, delta=0.1)






import warnings
from typing import Iterable
import time
import os

import numpy as np
import pandas as pd
import torch
import torch.linalg
from scipy.stats import spearmanr, hmean, gmean
from sklearn.utils.extmath import randomized_svd
import shap
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
from sklearn.datasets import load_svmlight_file
from pymoo.algorithms.soo.nonconvex.brkga import BRKGA
from pymoo.core.duplicate import ElementwiseDuplicateElimination
from pymoo.core.problem import ElementwiseProblem
from pymoo.optimize import minimize
from pymoo.termination.default import DefaultSingleObjectiveTermination
from pymoo.core.problem import StarmapParallelization
from multiprocessing.pool import ThreadPool


class ImportanceEvaluator:
    """
    Importance evaluator for VFL datasets
    """

    def __init__(self, sample_rate=0.01, seed=0):
        """
        :param sample_rate: [float] sample rate of the dataset to calculate Shaplley values
        :param seed: [int] random seed of sampling
        """
        self.sample_rate = sample_rate
        self.seed = seed
        np.random.seed(seed)  # This is for reproducibility of Permutation explainer

    @staticmethod
    def check_data(Xs):
        """
        Check if the data is valid for correlation evaluation.
        :param Xs: [List|Tuple] list of feature matrices of each party
        :return: [bool] True if the data is valid
        """
        n_features_on_party = [X.shape[1] for X in Xs]
        n_parties = len(n_features_on_party)
        assert n_parties >= 1, "ImportanceEvaluator only works for at least one party"
        assert all([X.shape[0] == Xs[0].shape[0] for X in Xs]), \
            "The number of samples should be the same for all parties"
        return n_features_on_party

    def evaluate_feature(self, X, model: callable, **kwargs):
        """
        Evaluate the importance of features in a dataset
        :param X: feature matrix
        :param model: [callable] model to be evaluated
        :param kwargs: [dict] other parameters for shap.explainers.Permutation
        :return: [np.ndarray] sum of importance on each party
        """
        # max_eval is 500 by default in sha.explainers.Permutation explainer
        # if the number of features is large, we need to increase max_evals
        if 'max_evals' not in kwargs and 2 * X.shape[1] + 1 > 500:
            kwargs['max_evals'] = 2 * X.shape[1] + 1

        # calculate Shapley values for each feature
        explainer = shap.explainers.Permutation(model, X, **kwargs)
        sample_size = int(self.sample_rate * X.shape[0])
        if sample_size <= 1:
            raise ValueError(f"Sample size 0 cannot produce meaningful Shapley values."
                             f" Please increase the sample rate.")
        X_sample = shap.sample(X, sample_size, random_state=self.seed)
        shap_values = explainer(X_sample).values

        importance_by_feature = np.sum(np.abs(shap_values), axis=0)
        assert importance_by_feature.shape[0] == X.shape[1], "The number of features should be the same"
        return importance_by_feature

    def evaluate(self, Xs, model: callable, **kwargs):
        """
        Evaluate the importance of features in VFL datasets
        :param Xs: [list] list of feature matrices
        :param model: [callable] model to be evaluated
        :param kwargs: [dict] other parameters for shap.explainers.Permutation
        :return: [np.ndarray] sum of importance on each party
        """
        n_features_on_party = self.check_data(Xs)
        X = np.concatenate(Xs, axis=1)

        # max_eval is 500 by default in sha.explainers.Permutation explainer
        # if the number of features is large, we need to increase max_evals
        if 'max_evals' not in kwargs and 2 * X.shape[1] + 1 > 500:
            kwargs['max_evals'] = 2 * X.shape[1] + 1

        # calculate Shapley values for each feature
        explainer = shap.explainers.Permutation(model, X, **kwargs)
        sample_size = int(self.sample_rate * X.shape[0])
        if sample_size <= 1:
            raise ValueError(f"Sample size 0 cannot produce meaningful Shapley values."
                             f" Please increase the sample rate.")
        X_sample = shap.sample(X, sample_size, random_state=self.seed)
        shap_values = explainer(X_sample).values

        # sum up the importance of each feature on each party
        party_cutoff = np.cumsum(n_features_on_party)
        party_cutoff = np.insert(party_cutoff, 0, 0)
        importance = np.zeros(len(n_features_on_party))
        for i in range(len(n_features_on_party)):
            party_shapley = np.sum(np.abs(shap_values[:, party_cutoff[i]:party_cutoff[i + 1]]), axis=1)
            importance[i] = np.mean(np.abs(party_shapley))
        assert len(importance) == len(n_features_on_party), \
            "The length of importance should be the same as the number of parties"
        return importance

    def evaluate_alpha(self, Xs=None, model: callable=None, scores=None, **kwargs):
        if scores is None:
            if Xs is None or model is None:
                raise ValueError(f"Since {scores=}, either scores or Xs and model should be provided to get scores.")
            scores = self.evaluate(Xs, model, **kwargs)
        num_parties = len(scores)
        scores = scores / np.sum(scores)  # normalize the scores
        score_var = np.var(scores)
        est_alpha = (num_parties - 1 - num_parties ** 2 * score_var) / (num_parties ** 3 * score_var)
        return est_alpha


class CorrelationEvaluator:
    """
    Correlation evaluator for VFL datasets
    """

    def __init__(self, corr_func='spearmanr', gpu_id=None, svd_algo='auto', n_jobs=1, **kwargs):
        """
        :param corr_func: [str] function to calculate the correlation between two features
        :param gamma: [float] weight of the inner-party correlation score
        :param gpu_id: [int] GPU id to use. If None, use CPU
        :param svd_algo: [str] algorithm to use for SVD. Should be one of {'auto', 'approx', 'exact'}
        :param kwargs: [dict] other parameters for pcor_singular
        """

        self.corr = None
        self.n_features_on_party = None
        self.gpu_id = gpu_id
        self.svd_algo = svd_algo

        assert self.svd_algo in ['auto', 'approx', 'exact'], "svd_algo should be auto, approx or exact"
        if self.gpu_id is not None:
            self.device = torch.device(f"cuda:{self.gpu_id}")

            if corr_func == "spearmanr":
                self.corr_func = self.spearmanr_pandas  # use CPU for now, a bug in GPU version
            elif corr_func == "spearmanr_pandas":
                self.corr_func = self.spearmanr_pandas
            elif corr_func == "pearson":
                self.corr_func = self.pearson_gpu
            else:
                raise NotImplementedError("corr_func should be in spearmanr, spearmanr_pandas or pearson")
        else:
            self.device = torch.device("cpu")
            if corr_func == "spearmanr":
                self.corr_func = self.spearmanr_pandas
            elif corr_func == "spearmanr_pandas":
                self.corr_func = self.spearmanr_pandas
            elif corr_func == "pearson":
                self.corr_func = self.pearson
            else:
                raise NotImplementedError("corr_func should be in spearmanr, spearmanr_pandas or pearson")

        print(f"CorrelationEvaluator uses {self.device}")
        self.pcor_kwargs = kwargs

        # parallel evaluator
        self.pool = ThreadPool(n_jobs)
        self.runner = StarmapParallelization(self.pool.starmap)

        self.min_icor = None
        self.max_icor = None

    def spearmanr_pandas(self, X):
        with np.errstate(divide='ignore', invalid='ignore'):
            corr = pd.DataFrame(X).corr(method='spearman').values
            # corr = spearmanr(X).correlation <=== BUG: this cannot calculate the correlation of constant features
        if np.isnan(corr).all():  # in case all features are constant
            corr = np.zeros((X.shape[1], X.shape[1]))
        else:
            corr = np.nan_to_num(corr, nan=0)
        if self.gpu_id is not None:
            corr = torch.from_numpy(corr).float().to(self.device)
        return corr

    def spearmanr(self, X):
        """
        Calculate the correlation matrix of X
        :param X: [np.ndarray] 2D data matrix. Size: n_samples * n_features
        :return: [np.ndarray] correlation matrix. Size: n_features * n_features
        """
        # When there are constant features in X. The correlation may be NaN, raise a warning "numpy ignore divide by
        # zero warning". We ignore this warning and replace NaN in corr with 0.
        with np.errstate(divide='ignore', invalid='ignore'):
            corr = spearmanr(X).correlation  # <=== BUG: this cannot calculate the correlation of constant features
        if np.isnan(corr).all():  # in case all features are constant
            raise ValueError("All features are constant, scipy have a BUG that cannot calculate the correlation.")
        else:
            corr = np.nan_to_num(corr, nan=0)
        if self.gpu_id is not None:
            corr = torch.from_numpy(corr).float().to(self.device)
        return corr

    def pearson(self, X):
        """
        Calculate the correlation matrix of X
        :param X: [np.ndarray] 2D data matrix. Size: n_samples * n_features
        :return: [np.ndarray] correlation matrix. Size: n_features * n_features
        """
        # When there are constant features in X. The correlation may be NaN, raise a warning "numpy ignore divide by
        # zero warning". We ignore this warning and replace NaN in corr with 0.
        with np.errstate(divide='ignore', invalid='ignore'):
            corr = np.corrcoef(X, rowvar=False)
        if np.isnan(corr).all():  # in case all features are constant
            corr = np.zeros((X.shape[1], X.shape[1]))
        else:
            corr = np.nan_to_num(corr, nan=0)
        if self.gpu_id is not None:
            corr = torch.from_numpy(corr).float().to(self.device)
        return corr

    def pearson_gpu(self, X):
        """
        Calculate the correlation matrix of X
        :param X: [np.ndarray] 2D data matrix. Size: n_samples * n_features
        :return: [np.ndarray] correlation matrix. Size: n_features * n_features
        """
        # When there are constant features in X. The correlation may be NaN, raise a warning "numpy ignore divide by
        # zero warning". We ignore this warning and replace NaN in corr with 0.
        X = torch.from_numpy(X).float().to(self.device)
        corr = torch.corrcoef(X.T)  # columns are variables
        corr = torch.nan_to_num(corr, nan=0)
        return corr

    @staticmethod
    def pcor_singular_naive(corr):
        """
        Calculate the overall correlation score of a correlation matrix using the variance of singular values.
        This is a naive implementation of the method that calculates all singular values. This is usually 2 to
        3 times slower than pcor_singular_exact().
        This is an example of corr
        [ v1 v1 ]
        [ v1 v1 ]
        :param corr:
        :return:
        """
        # start_time = time.time()
        assert np.isnan(corr).any() == False, "NaN values should be replaced with 0"

        # d = corr.shape[0]
        singular_values = np.linalg.svd(corr)[1]
        score = np.std(singular_values, ddof=1) / np.sqrt(min(corr.shape))

        # end_time = time.time()
        # print(f"Time for calculating the correlation score: {end_time - start_time}")
        return float(score.item())

    @staticmethod
    def pcor_singular_exact(corr):
        """
        [Exact] Calculate the overall correlation score of a correlation matrix using the variance of singular values.
        Using the definition of std: std = sqrt(E[X^2] - E[X]^2). E[X^2] can be calculated by trace(X^TX), which is
        square of Frobenius norm. E[X] is known as nuclear norm.
        This is an example of corr
        [ v1 v1 ]
        [ v1 v1 ]
        :param corr:
        :return:
        """
        assert np.isnan(corr).any() == False, "NaN values should be replaced with 0"

        d = min(corr.shape[0], corr.shape[1])
        vals = np.linalg.svd(corr, compute_uv=False)
        if len(vals) <= 1:
            score = np.array([0])  # less than one feature, we define std as 0
        else:
            score = np.std(vals, ddof=1) / np.sqrt(d)

        return float(score.item())

    def pcor_singular_exact_gpu(self, corr: torch.Tensor):
        """
        [Exact] Calculate the overall correlation score of a correlation matrix using the variance of singular values.
        Using the definition of std: std = sqrt(E[X^2] - E[X]^2). E[X^2] can be calculated by trace(X^TX), which is
        square of Frobenius norm. E[X] is known as nuclear norm.
        This is an example of corr
        [ v1 v1 ]
        [ v1 v1 ]
        :param corr:
        :return:
        """
        singular_values = torch.linalg.svdvals(corr)
        d = min(corr.shape)
        if len(singular_values) <= 1:
            score = torch.tensor([0])  # less than one feature, we define std as 0
        else:
            score = torch.std(singular_values) / np.sqrt(d)

        return float(score.item())

    @staticmethod
    def pcor_singular_approx(corr, n_components=400, n_oversamples=10, n_iter=4, random_state=0):
        """
        [Approximate] Calculate the overall correlation score of a correlation matrix using the variance of singular
        values. This function uses randomized SVD to approximate the singular values. It is much faster than the exact
        in high-dimensional data at the cost of some accuracy.
        This is an example of corr
        [ v1 v1 ]
        [ v1 v1 ]
        :param corr:
        :param n_components: [int] number of singular values to return
        :param n_oversamples: [int] usually set to 2*k-n_components when k is the effective rank of the matrix
        :param n_iter: [int] number of power iterations
        :param random_state: [int] random seed
        :param gpu_id: [int] GPU id. If None, use CPU
        :return:
        """
        # start_time = time.time()
        if n_components > min(corr.shape):
            n_components = min(corr.shape)

        assert np.isnan(corr).any() == False, "NaN values should be replaced with 0"
        _, singular_values, _ = randomized_svd(corr, n_components=n_components, n_oversamples=n_oversamples,
                                               n_iter=n_iter,
                                               random_state=random_state)
        singular_shape = min(corr.shape)
        s_append_zero = np.concatenate((singular_values, np.zeros(singular_shape - singular_values.shape[0])))

        if len(s_append_zero) <= 1:
            score = np.array([0])
        else:
            score = np.std(s_append_zero, ddof=1) / np.sqrt(singular_shape)

        # end_time = time.time()
        # print(f"Time for calculating the correlation score: {end_time - start_time}")
        return float(score.item())

    @staticmethod
    def pcor_singular_approx_gpu(corr: torch.Tensor, n_components=400, n_iter=4):
        """
        [Approximate] Calculate the overall correlation score of a correlation matrix using the variance of singular
        values. This function uses randomized SVD to approximate the singular values. It invokes torch.svd_lowrank() to
        calculate the singular values on GPU.
        This is an example of corr
        [ v1 v1 ]
        [ v1 v1 ]
        :param corr: [torch.Tensor] 2D data matrix. Size: n_features * n_features. This matrix should be already on GPU.
        :param n_components: [int] number of singular values to return
        :param n_iter: [int] number of power iterations
        :return: [float] correlation score
        """

        if n_components > min(corr.shape):
            n_components = min(corr.shape)

        _, singular_values, _ = torch.svd_lowrank(corr, q=n_components, niter=n_iter)
        singular_shape = min(corr.shape)
        s_append_zero = torch.concatenate((singular_values,
                                           torch.zeros(singular_shape - singular_values.shape[0]).to(corr.device)))

        if len(s_append_zero) <= 1:
            score = torch.tensor([0])
        else:
            score = torch.std(s_append_zero) / np.sqrt(singular_shape)

        return float(score.item())

    def pcor_singular(self, corr, **kwargs):
        """
        Calculat the std of the singular values of corr matrix.
        :param corr: [np.ndarray] correlation matrix
        :param kwargs:

        :param self.svd_algo: [str] algorithm to calculate the overall correlation score of a correlation matrix
                        - 'auto': automatically choose the algorithm based on the size of the correlation matrix
                                  if the size is smaller than 200, use 'exact', otherwise use 'approx'
                        - 'exact': calculate the exact singular values
                        - 'approx': calculate the approximate singular values
        :return:
        """
        # merge kwargs with self.kwargs, and overwrite self.kwargs if there is a conflict
        kwargs = self.pcor_kwargs | kwargs
        kwargs.pop('method', None)  # remove the unused self.kwargs['method']

        if self.svd_algo == 'auto':
            if min(corr.shape) < 100:
                if self.gpu_id is not None:
                    return self.pcor_singular_exact_gpu(corr)
                else:
                    return CorrelationEvaluator.pcor_singular_exact(corr)
            else:
                if self.gpu_id is not None:
                    return self.pcor_singular_approx_gpu(corr, **kwargs)
                else:
                    return CorrelationEvaluator.pcor_singular_approx(corr, **kwargs)
        elif self.svd_algo == 'exact':
            if self.gpu_id is not None:
                return self.pcor_singular_exact_gpu(corr)
            else:
                return CorrelationEvaluator.pcor_singular_exact(corr)
        elif self.svd_algo == 'approx':
            if self.gpu_id is not None:
                return self.pcor_singular_approx_gpu(corr, **kwargs)
            else:
                return CorrelationEvaluator.pcor_singular_approx(corr, **kwargs)
        else:
            raise ValueError(f"Unknown algorithm: {self.svd_algo}")

    def _get_inner_and_inter_corr(self, corr, n_features_on_party, symmetric=True):
        """
        Calculate the inner-party and inter-party correlation matrices.
        :param corr: [np.ndarray] correlation matrix
        :param n_features_on_party: [list] number of features on each party
        :return: [2D np.ndarray] correlation scores (size: number of parties x number of parties)
        """
        n_parties = len(n_features_on_party)
        assert sum(n_features_on_party) == corr.shape[0] == corr.shape[1], \
            f"The number of features on each party should be the same as the size of the correlation matrix," \
            "but got {sum(n_features_on_party)} != {corr.shape[0]} != {corr.shape[1]}"
        corr_cut_points = np.cumsum(n_features_on_party)
        corr_cut_points = np.insert(corr_cut_points, 0, 0)

        pcors = np.zeros((n_parties, n_parties))
        for i in range(n_parties):
            for j in range(n_parties):
                start_i = corr_cut_points[i].item()
                end_i = corr_cut_points[i + 1].item()
                start_j = corr_cut_points[j].item()
                end_j = corr_cut_points[j + 1].item()

                if symmetric and j > i:
                    continue

                pcors[i][j] = self.pcor_singular(corr[start_i:end_i, start_j:end_j])

        if symmetric:
            for i in range(n_parties):
                for j in range(i + 1, n_parties):
                    pcors[i][j] = pcors[j][i]
        return pcors

    def _get_inter_corr(self, corr, n_features_on_party, symmetric=True):
        """
        Calculate the inter-party correlation matrices.
        :param corr: [np.ndarray|torch.Tensor] correlation matrix
        :param n_features_on_party: [list] number of features on each party
        :param symmetric: [bool] whether the inter-party correlation matrix is symmetric. If True, only the upper
                            triangular part of the matrix is calculated.
        :return: [2D np.ndarray] inter-party correlation scores (size: number of parties x (number of parties - 1))
        """
        n_parties = len(n_features_on_party)
        assert sum(n_features_on_party) == corr.shape[0] == corr.shape[1], \
            f"The number of features on each party should be the same as the size of the correlation matrix," \
            "but got {sum(n_features_on_party)} != {corr.shape[0]} != {corr.shape[1]}"
        corr_cut_points = np.cumsum(n_features_on_party)
        corr_cut_points = np.insert(corr_cut_points, 0, 0)

        inter_pcors = []
        for i in range(n_parties):
            for j in range(n_parties):
                start_i = corr_cut_points[i].item()
                end_i = corr_cut_points[i + 1].item()
                start_j = corr_cut_points[j].item()
                end_j = corr_cut_points[j + 1].item()
                if symmetric:
                    save = i < j
                else:
                    save = i != j
                if save:
                    inter_pcors.append(self.pcor_singular(corr[start_i:end_i, start_j:end_j]))

        return np.array(inter_pcors)

    def overall_corr_score(self, corr, n_features_on_party):
        """
        Calculate the correlation score of a correlation matrix. This is a three-party example of corr:
        [ v1 v1 .  .  .  .  ]
        [ v1 v1 .  .  .  .  ]
        [ .  .  v2 v2 .  .  ]
        [ .  .  v2 v2 .  .  ]
        [ .  .  .  .  v3 v3 ]
        [ .  .  .  .  v3 v3 ]
        v1, v2, v3 represent inner-party correlation matrices of each party. This function evaluates the
        arithmetic mean of differences between the inter-party correlation scores and the inner-party correlation
        scores for each party.
        :param corr: [np.ndarray|torch.Tensor] correlation matrix
        :param n_features_on_party: [list] number of features on each party
        :return: [float] correlation score
        """
        # inter_pcors = self._get_inter_corr(corr, n_features_on_party, pcor_func)
        pcors = self._get_inner_and_inter_corr(corr, n_features_on_party)
        N = pcors.shape[0]

        # calculate the absolute difference between the inner-party correlation scores and the inter-party correlation
        # for each party
        diffs = np.zeros(N)
        for i in range(N):
            for j in range(N):
                if i != j:
                    diffs[i] += (pcors[j][i] - pcors[i][i])  # in range [-1, 1]
            diffs[i] /= (N - 1)
        return np.mean(diffs)

    @staticmethod
    def check_data(Xs):
        """
        Check if the data is valid for correlation evaluation.
        :param Xs: [List|Tuple] list of feature matrices of each party
        :return: [bool] True if the data is valid
        """
        n_features_on_party = [X.shape[1] for X in Xs]
        n_parties = len(n_features_on_party)
        assert n_parties >= 1, "The number of parties should be at least 1"
        assert all([X.shape[0] == Xs[0].shape[0] for X in Xs]), \
            "The number of samples should be the same for all parties"
        return n_features_on_party

    def fit_evaluate(self, Xs):
        """
        Evaluate the correlation score of a vertical federated learning dataset.
        :param Xs: [List|Tuple] list of feature matrices of each party
        :return: [float] correlation score
        """
        self.fit(Xs)
        return self.evaluate(Xs)

    def fit(self, Xs):
        """
        Fit the correlation matrix of a vertical federated learning dataset.
        :param Xs: [List|Tuple] list of feature matrices of each party
        """
        self.n_features_on_party = self.check_data(Xs)
        if Xs[0].shape[0] <= 1:
            raise ValueError("The number of samples should be greater than 1")
        Xs = list(Xs)
        if torch.is_tensor(Xs[0]):
            self.corr = self.corr_func(torch.cat(Xs, dim=1))
        elif isinstance(Xs[0], np.ndarray):
            self.corr = self.corr_func(np.concatenate(Xs, axis=1))
        else:
            raise ValueError(f"Xs should be either np.ndarray or torch.Tensor, but got {type(Xs[0])}")

        if torch.is_tensor(self.corr):
            self.corr = torch.nan_to_num(self.corr, nan=0)
        else:
            self.corr = np.nan_to_num(self.corr, nan=0)

    class DuplicationElimination(ElementwiseDuplicateElimination):
        def is_equal(self, perm_a, perm_b):
            return perm_a.get("hash") == perm_b.get("hash")

    # Nested class for BRKGA solver: max-pcor problem definition
    class CorrMaxProblem(ElementwiseProblem):
        def __init__(self, corr, n_features_on_party, evaluator=None, runner=None):
            super().__init__(n_var=corr.shape[1], n_obj=1, n_constr=0, xl=-1, xu=1, elementwise_runner=runner)
            self.corr = corr
            self.n_features_on_party = n_features_on_party
            self.evaluator = evaluator

        # @cachetools.cached(cache=cachetools.TTLCache(maxsize=1000, ttl=60),
        #                    key=lambda self, corr, order: hash(tuple(
        #                        CorrelationSplitter.sort_order_by_party(order, self.n_features_on_party))))
        def corr_score(self, corr, order):
            corr_perm = corr[order, :][:, order]
            return self.evaluator.overall_corr_score(corr_perm, self.n_features_on_party)

        def _evaluate(self, x, out, *args, **kwargs):
            order = np.argsort(x)
            out['F'] = -self.corr_score(self.corr, order)
            out['order'] = order
            # sorted_order_by_party = CorrelationSplitter.sort_order_by_party(order, self.n_features_on_party)
            out['hash'] = hash(tuple(order))

    # Nested class for BRKGA solver: min-pcor problem definition
    class CorrMinProblem(ElementwiseProblem):
        def __init__(self, corr, n_features_on_party, evaluator=None, runner=None):
            super().__init__(n_var=corr.shape[1], n_obj=1, n_constr=0, xl=-1, xu=1, elementwise_runner=runner)
            self.corr = corr
            self.n_features_on_party = n_features_on_party
            self.evaluator = evaluator

        def corr_score(self, corr, order):
            corr_perm = corr[order, :][:, order]
            return self.evaluator.overall_corr_score(corr_perm, self.n_features_on_party)

        def _evaluate(self, x, out, *args, **kwargs):
            order = np.argsort(x)
            out['F'] = self.corr_score(self.corr, order)
            out['order'] = order
            # sorted_order_by_party = CorrelationSplitter.sort_order_by_party(order, self.n_features_on_party)
            out['hash'] = hash(tuple(order))

    def fit_min_max(self, n_elites=200, n_offsprings=700, n_mutants=100, n_gen=100, bias=0.7, verbose=False,
                    seed=None, **kwargs):
        """
        Find the min-max value of Icor for all the possible feature split
        :param X: [np.ndarray] global data set
        :return: [tuple(float, float)] min-max value of Icor
        """
        # self.evaluator.corr = self.evaluator.corr_func(X)
        # self.evaluator.n_features_on_party = self.split_num_features_equal(X.shape[1], self.num_parties)
        if self.corr is None:
            raise ValueError("Data matrix X is not provided, and self.corr is not calculated yet."
                             "Please either provide X or call fit() first.")

        algorithm = BRKGA(
            n_elites=n_elites,
            n_offsprings=n_offsprings,
            n_mutants=n_mutants,
            bias=bias,
            eliminate_duplicates=self.DuplicationElimination(),
        )

        # calculate the min and max pcor
        if verbose:
            print("Calculating the min pcor of the overall correlation score...")
        res_min = minimize(
            self.CorrMinProblem(self.corr, self.n_features_on_party, evaluator=self,
                                runner=self.runner),
            algorithm,
            ('n_gen', n_gen),
            seed=seed,
            verbose=verbose,
        )
        self.min_icor = res_min.F[0]
        print(f"min_icor: {self.min_icor}")
        if verbose:
            print("Calculating the max icor of the overall correlation score...")
        res_max = minimize(
            self.CorrMaxProblem(self.corr, self.n_features_on_party, evaluator=self,
                                runner=self.runner),
            algorithm,
            ('n_gen', n_gen),
            seed=seed,
            verbose=verbose,
        )
        self.max_icor = -res_max.F[0]
        print(f"max_icor: {self.max_icor}")

    def evaluate(self, Xs=None):
        """
        Evaluate the correlation score of a vertical federated learning dataset with self.corr.
        :return: [float] correlation score
        """
        if self.corr is None:
            raise ValueError("Please call fit() or fit_evaluate() first to calculate the correlation matrix.")
        if Xs is not None:
            # If Xs is provided, update self.n_features_on_party
            self.n_features_on_party = self.check_data(Xs)
        return self.overall_corr_score(self.corr, self.n_features_on_party)

    def evaluate_beta(self, Xs=None, clip=True) -> float:
        if self.corr is None:
            raise ValueError("Please call fit() or fit_evaluate() first to calculate the correlation matrix.")

        if self.min_icor is None or self.max_icor is None:
            # evaluate the min and max icor
            self.fit_min_max()

        score = self.evaluate(Xs)  # when Xs is None, self.corr and self.n_features_on_party should be already set
        beta = (score - self.min_icor) / (self.max_icor - self.min_icor)
        if clip:
            return np.clip(beta, 0, 1).item()
        else:
            return beta

    def visualize(self, save_path=None, value=None, cmap='cividis', map_func=np.abs, fontsize=16, title_size=None):
        """
        Visualize the correlation matrix with the latest self.n_features_on_party and self.corr.
        :param map_func: [callable|None|str] function to map the correlation matrix. If None, the correlation matrix will
        be used directly.
        :param cmap: [str] color map for the figure
        :param save_path: [str|None] path to save the figure. If None, the figure will be shown.
        :param value: [float|None] The overall correlation score to be shown on the figure. If None, the score will not
        be shown.
        """
        if save_path and not os.path.exists(os.path.dirname(save_path)):
            os.makedirs(os.path.dirname(save_path))

        if self.corr is None:
            raise ValueError("Please call fit() or fit_evaluate() first to calculate the correlation matrix.")
        if type(self.corr) == torch.Tensor:
            corr = self.corr.cpu().numpy()
        else:
            corr = self.corr

        if title_size is None:
            title_size = fontsize

        if map_func is None:
            map_corr = corr
        elif isinstance(map_func, str):
            # corr is in [-1, 1]
            if map_func == 'sigmoid':
                map_corr = 1 / (1 + np.exp(-corr))
            elif map_func == 'tanh':
                map_corr = np.tanh(corr)
            elif map_func == 'relu':
                map_corr = np.maximum(0, corr)
            elif map_func == 'log':
                map_corr = np.log(corr + 1)
            elif map_func == 'exp':
                map_corr = np.exp(corr)
            elif map_func == 'abs':
                map_corr = np.abs(corr)
            else:
                raise ValueError(f"Unknown map function {map_func} (str)")
        elif callable(map_func):
            map_corr = map_func(corr)
        else:
            raise ValueError(f"Unknown map function {map_func} (callable)")

        plt.rc('font', size=fontsize)

        fig, ax = plt.subplots(figsize=(10, 9))
        plt.imshow(map_corr, cmap=cmap)
        plt.colorbar()

        # add the axis to indicate the party
        n_features_on_party = np.asarray(self.n_features_on_party)
        n_parties = len(n_features_on_party)
        xtick_major = np.cumsum([0] + self.n_features_on_party)
        xtick_minor = (xtick_major[1:] + xtick_major[:-1]) / 2
        xtick_labels = [f"Party {i + 1}" for i in range(n_parties)]

        # add lines to separate the parties
        for i in range(n_parties - 1):
            plt.axvline(x=xtick_major[i + 1] - 0.5, color='red', linewidth=2)
            plt.axhline(y=xtick_major[i + 1] - 0.5, color='red', linewidth=2)

        ax.xaxis.set_major_locator(ticker.FixedLocator(xtick_major))
        ax.yaxis.set_major_locator(ticker.FixedLocator(xtick_major))
        ax.xaxis.set_minor_locator(ticker.FixedLocator(xtick_minor))
        ax.yaxis.set_minor_locator(ticker.FixedLocator(xtick_minor))
        ax.set_xticklabels(xtick_labels, minor=True)
        ax.set_yticklabels(xtick_labels, minor=True)
        ax.tick_params(axis='both', which='minor', length=0)
        ax.tick_params(axis='x', which='minor', rotation=90)
        ax.tick_params(axis='x', which='major', rotation=90)

        fig.tight_layout()

        # move title up
        if value is not None:
            plt.title(f"Correlation matrix (Icor={value:.2f})", y=1.05, fontsize=title_size)
        else:
            plt.title("Correlation matrix", y=1.05, fontsize=title_size)
        if save_path is None:
            plt.show()
        else:
            plt.savefig(save_path)
        plt.close()


if __name__ == '__main__':
    X, y = load_svmlight_file("data/real/vehicle/processed/vehicle.libsvm")
    X = X.toarray()

    Xs = [X[:, :50], X[:, 50:]]
    evaluator = CorrelationEvaluator(gpu_id=0)
    score = evaluator.fit_evaluate(Xs)
    evaluator.visualize(value=score, cmap="afmhot")



import unittest

from collections import defaultdict
import numpy as np
import xgboost as xgb
from sklearn.datasets import make_classification
from sklearn.linear_model import LinearRegression

from src.vertibench.Evaluator import ImportanceEvaluator, CorrelationEvaluator


class TestAlphaEvaluator(unittest.TestCase):
    def setUp(self):
        self.X, self.y = make_classification(n_samples=1000, n_features=100, n_informative=50, n_redundant=0,
                                             random_state=0)
    def test_evaluate_alpha(self):
        # Create a model
        model = xgb.XGBClassifier()
        model.fit(self.X, self.y)

        # Create an evaluator
        evaluator = ImportanceEvaluator(sample_rate=0.1)
        scores = evaluator.evaluate_feature(self.X, model.predict)

        split_ratios = [0.1, 0.2, 0.3, 0.4, 0.5]
        alpha1s = []
        alpha2s = []
        for split_ratio in split_ratios:
            with self.subTest(split_ratio=split_ratio):
                # Split the data
                l_idx = int(split_ratio * self.X.shape[1])
                Xs = [self.X[:, :l_idx], self.X[:, l_idx:]]
                ys = [self.y, self.y]
                l_score, r_score = np.sum(scores[:l_idx]) / np.sum(scores), np.sum(scores[l_idx:]) / np.sum(scores)

                # evaluate feature
                scores_per_party = evaluator.evaluate(Xs, model.predict)
                self.assertEqual(len(scores_per_party), 2)
                self.assertTrue(np.allclose(scores_per_party[0], l_score, atol=0.2),
                                msg=f"{scores_per_party[0]} != {l_score}")
                self.assertTrue(np.allclose(scores_per_party[1], r_score, atol=0.2),
                                msg=f"{scores_per_party[1]} != {r_score}")

                # Evaluate alpha
                alpha1 = evaluator.evaluate_alpha(scores=scores_per_party)
                alpha2 = evaluator.evaluate_alpha(Xs=Xs, model=model.predict)

                alpha1s.append(alpha1)
                alpha2s.append(alpha2)

        with self.subTest("overall"):
            # mse between two alpha evaluations should be small
            diff_scale = np.abs(np.array(alpha1s) - np.array(alpha2s)) / alpha2s
            self.assertLess(np.mean(diff_scale), 0.3)

            # Check monotonicity, alphas should be generally decreasing
            # use linear regression to check monotonicity
            lr = LinearRegression()
            lr.fit(np.array(split_ratios).reshape(-1, 1), np.array(alpha1s).reshape(-1, 1))
            self.assertGreater(lr.coef_[0][0], 0)



from numbers import Real
import warnings
import abc

import numpy as np
import torch
import torch.linalg
from pymoo.algorithms.soo.nonconvex.brkga import BRKGA
from pymoo.core.duplicate import ElementwiseDuplicateElimination
from pymoo.core.problem import ElementwiseProblem
from pymoo.optimize import minimize
from pymoo.termination.default import DefaultSingleObjectiveTermination
from pymoo.core.problem import StarmapParallelization
from multiprocessing.pool import ThreadPool

from .Evaluator import CorrelationEvaluator


class Splitter(abc.ABC):
    def __init__(self, num_parties):
        self.num_parties = num_parties

    @abc.abstractmethod
    def split_indices(self, *args, **kwargs):
        """
        Split the indices of X
        :param X: [np.ndarray] 2D dataset
        :return: [list] list of indices of each party
        """
        pass

    def split(self, *Xs, indices=None, allow_empty_party=False, fill=None):
        assert len(Xs) > 0, "At least one dataset should be given"
        ans = []

        # calculate the indices for each party for all datasets
        if indices is None:
            allX = np.concatenate(Xs, axis=0)
            party_to_feature = self.split_indices(allX, allow_empty_party=allow_empty_party)
        else:
            party_to_feature = indices

        # split each dataset
        for X in Xs:
            X_split = []
            for i in range(self.num_parties):
                selected = party_to_feature[i]  # selected column_ids
                if fill is not None:
                    X_party_i = np.full_like(X, fill)
                    X_party_i[:, selected] = X[:, selected]
                else:
                    X_party_i = X[:, selected]
                X_split.append(X_party_i)
            ans.append(X_split)

        if len(ans) == 1:
            return ans[0]
        else:
            return ans


class ImportanceSplitter(Splitter):
    def __init__(self, num_parties, weights=1, seed=None):
        """
        Split a 2D dataset by feature importance under dirichlet distribution (assuming the features are independent).
        :param num_parties: [int] number of parties
        :param weights: [int | list with size num_parties]
                        If weights is an int, the weight of each party is the same.
                        If weights is an array, the weight of each party is the corresponding element in the array.
                        The weights indicate the expected sum of feature importance of each party.
                        Meanwhile, larger weights mean less bias on the feature importance.
        :param seed: [int] random seed
        """
        super().__init__(num_parties)
        self.weights = weights
        self.seed = seed
        np.random.seed(seed)
        if isinstance(self.weights, Real):
            self.weights = [self.weights for _ in range(self.num_parties)]

        self.check_params()

    def check_params(self):
        """
        Check if the parameters are valid
        """
        if len(self.weights) != self.num_parties:
            raise ValueError("The length of weights should equal to the number of parties")
        if not all([weight > 0 for weight in self.weights]):
            raise ValueError("The weights should be positive")
        if self.num_parties < 2:
            raise ValueError("The number of parties should be greater than 1")

    @staticmethod
    def dirichlet(alpha):
        """
        Generate a random sample from a Dirichlet distribution using beta distribution. This function can
        work with small alpha values.
        :param alpha: [float] the parameter of the symmetric Dirichlet distribution
        :return: [np.ndarray] the generated sample
        """
        xs = [np.random.beta(alpha[0], sum(alpha[1:]))]
        for i in range(1, len(alpha) - 1):
            phi = np.random.beta(alpha[i], sum(alpha[i + 1:]))
            xs.append((1 - sum(xs)) * phi)
        xs.append(1 - sum(xs))
        return np.array(xs)

    def split_indices(self, X, allow_empty_party=False):
        """
        Split the indices of X by feature importance.
        :param allow_empty_party: [bool] whether to allow parties with zero features
        :param X: [np.ndarray] 2D dataset
        :return: [list] list of indices of each party
        """
        # Generate the probabilities of being assigned to each party
        # All the features share the same ratio
        # probs = np.random.dirichlet(self.weights)     # has bug with small weights
        # if np.isnan(probs).any():
        #     probs = self.dirichlet(self.weights)      # use beta distribution instead, slower but more accurate
        probs = self.dirichlet(self.weights)

        if np.sum(probs) < 1 - 1e-6:
            # very small weights may cause all probs to be 0
            # in this case, assign all features to a random party
            assert np.isclose(sum(probs), 0)
            target_party = np.random.randint(self.num_parties)
            probs[target_party] = 1

        if allow_empty_party:
            # Assign each feature to a party
            party_to_feature = [[] for _ in range(self.num_parties)]
            party_ids = np.random.choice(self.num_parties, size=X.shape[1], p=probs)
            for feature_id in range(X.shape[1]):
                party_id = party_ids[feature_id]
                party_to_feature[party_id].append(feature_id)
        else:
            # uniform-randomly select one feature for each party
            preassigned_feature_ids = np.random.choice(X.shape[1], size=self.num_parties, replace=False)
            party_to_feature = [[feature_id] for feature_id in preassigned_feature_ids]

            # Assign the remaining features to the parties
            preassigned_feature_id_set = set(preassigned_feature_ids)
            if not np.isclose(sum(probs), 1):
                raise ValueError(f"The sum of probs should be equal to 1, got {probs} from weights {self.weights}")
            party_ids = np.random.choice(self.num_parties, size=X.shape[1], p=probs)
            for feature_id in range(X.shape[1]):
                if feature_id not in preassigned_feature_id_set:
                    party_id = party_ids[feature_id]
                    party_to_feature[party_id].append(feature_id)

            # no empty party
            assert np.count_nonzero([len(party) for party in party_to_feature]) == self.num_parties, \
                "There should be no empty party"
            assert np.sum([len(party) for party in party_to_feature]) == X.shape[1], \
                "The number of features should be the same as the number of columns of X"

        return party_to_feature


class CorrelationSplitter(Splitter):

    def __init__(self, num_parties: int, evaluator: CorrelationEvaluator = None, seed=None, gpu_id=None, n_jobs=1):
        """
        Split a 2D dataset by feature correlation (assuming the features are equally important).
        :param num_parties: [int] number of parties
        :param evaluator: [CorrelationEvaluator] the evaluator to evaluate the correlation
        :param seed: [int] random seed
        :param gpu_id: [int] GPU id
        """
        super().__init__(num_parties)
        self.evaluator = evaluator
        if evaluator is None:
            self.evaluator = CorrelationEvaluator(gpu_id=gpu_id)
        self.seed = seed
        self.gpu_id = gpu_id
        if self.gpu_id is not None:
            assert self.evaluator.gpu_id == self.gpu_id, \
                "The gpu_id of the evaluator should be the same as the gpu_id of the splitter"
            self.device = torch.device("cuda:{}".format(self.gpu_id))
        else:
            self.device = torch.device("cpu")
            warnings.warn("The device is set to CPU. This may cause the performance to be slow.")

        self.pool = ThreadPool(n_jobs)
        self.runner = StarmapParallelization(self.pool.starmap)

        # split result of the last call of fit()
        # self.corr = None      # use evaluator.corr instead
        # self.n_features_on_party = None   # use evaluator.n_features_on_party instead
        self.min_icor = None
        self.max_icor = None

        # best split result of all calls of split()
        self.best_icor = None
        self.best_error = None
        self.best_feature_per_party = None
        self.best_permutation = None

    @staticmethod
    def sort_order_by_party(order, n_features_on_party):
        order_cut_points = np.cumsum(n_features_on_party)
        order_cut_points = np.insert(order_cut_points, 0, 0)
        sorted_order_by_party = []
        for i in range(1, len(order_cut_points)):
            sorted_order_by_party.append(tuple(sorted(order[order_cut_points[i - 1]:order_cut_points[i]])))
        return sorted_order_by_party

    # Nested class for BRKGA solver: duplicate elimination of permutations
    class DuplicationElimination(ElementwiseDuplicateElimination):
        def is_equal(self, perm_a, perm_b):
            return perm_a.get("hash") == perm_b.get("hash")

    # Nested class for BRKGA solver: best-matched-icor problem definition
    class CorrBestMatchProblem(ElementwiseProblem):
        def __init__(self, corr, n_features_on_party, beta, min_icor, max_icor, evaluator: CorrelationEvaluator = None,
                     runner=None):
            super().__init__(n_var=corr.shape[1], n_obj=1, n_constr=0, xl=-1, xu=1, elementwise_runner=runner)
            assert min_icor < max_icor, f"min_icor {min_icor} should be smaller than max_icor {max_icor}"
            self.corr = corr
            self.n_features_on_party = n_features_on_party
            self.beta = beta
            self.max_icor = max_icor
            self.min_icor = min_icor
            self.evaluator = evaluator

        # @cachetools.cached(cache=cachetools.TTLCache(maxsize=10000, ttl=60),
        #                    key=lambda self, corr, order: hash(tuple(
        #                        CorrelationSplitter.sort_order_by_party(order, self.n_features_on_party))))
        def corr_score(self, corr, order):
            corr_perm = corr[order, :][:, order]
            return self.evaluator.overall_corr_score(corr_perm, self.n_features_on_party)

        def _evaluate(self, x, out, *args, **kwargs):
            order = np.argsort(x)
            icor = self.corr_score(self.corr, order)
            target_icor = self.beta * self.max_icor + (1 - self.beta) * self.min_icor

            out['F'] = np.abs(icor - target_icor)
            out['icor'] = icor
            out['order'] = order
            # sorted_order_by_party = CorrelationSplitter.sort_order_by_party(order, self.n_features_on_party)
            out['hash'] = hash(tuple(order))

    def check_fit_data(self):
        """
        Check if the required members are calculated by fit()
        """
        assert self.evaluator.corr is not None, "self.evaluator.corr is None. Please call fit() first."
        assert self.evaluator.n_features_on_party is not None, "self.evaluator.n_features_on_party is None. Please call fit() first."
        assert self.min_icor is not None, "self.min_icor is None. Please call fit() first."
        assert self.max_icor is not None, "self.max_icor is None. Please call fit() first."

    def fit(self, X, **kwargs):
        """
        Calculate the min and max icor of the overall correlation score.
        Required parameters:
        :param X: [np.ndarray] 2D dataset

        Optional parameters: (BRKGA parameters)
        :param n_elites: (int) number of elites in BRKGA
        :param n_offsprings: (int) number of offsprings in BRKGA
        :param n_mutants: (int) number of mutants in BRKGA
        :param n_gen: (int) number of generations in BRKGA
        :param bias: (float) bias of BRKGA
        :param verbose: (bool) whether to print the progress
        :param kwargs: other unused args
        """
        Xs = np.array_split(X, self.num_parties, axis=1)
        self.evaluator.fit(Xs)
        self.evaluator.fit_min_max(**kwargs)
        self.min_icor = self.evaluator.min_icor
        self.max_icor = self.evaluator.max_icor

    def split_indices(self, X, n_elites=20, n_offsprings=70, n_mutants=10, n_gen=100, bias=0.7, verbose=False,
              beta=0.5, term_tol=1e-4, term_period=10):
        """
        Use BRKGA to find the best order of features that minimizes the difference between the mean of icor and the
        target. split() assumes that the min and max icor have been calculated by fit().
        Required parameters:
        :param X: [np.ndarray] 2D dataset

        Optional parameters: (BRKGA parameters)
        :param n_elites: (int) number of elites in BRKGA
        :param n_offsprings: (int) number of offsprings in BRKGA
        :param n_mutants: (int) number of mutants in BRKGA
        :param n_gen: (int) number of generations in BRKGA
        :param bias: (float) bias of BRKGA
        :param verbose: (bool) whether to print the progress
        :param beta: [float] the tightness of inner-party correlation. Larger beta means more inner-party correlation
                             and less inter-party correlation.
        :param term_tol: (float) If out['F'] < term_tol after term_period generations, the algorithm terminates.
        :param term_period: (int) Check the termination condition every term_period generations

        :return: (np.ndarray) indices of features in the order of importance
        """
        self.check_fit_data()
        if not (0 <= beta <= 1):
            raise ValueError(f"beta should be in [0, 1], got {beta}")

        # termination by number of generations or the error is less than 1e-6
        termination = DefaultSingleObjectiveTermination(ftol=term_tol, n_max_gen=n_gen, period=term_period)
        algorithm = BRKGA(
            n_elites=n_elites,
            n_offsprings=n_offsprings,
            n_mutants=n_mutants,
            bias=bias,
            eliminate_duplicates=self.DuplicationElimination(),
        )

        # find the best permutation order that makes the icor closest to the target icor
        # target_icor = beta * max_icor + (1 - beta) * min_icor
        res_beta = minimize(
            self.CorrBestMatchProblem(self.evaluator.corr, self.evaluator.n_features_on_party, beta, self.min_icor,
                                      self.max_icor,
                                      evaluator=self.evaluator, runner=self.runner),
            algorithm,
            termination,
            seed=self.seed,
            verbose=verbose,
        )
        self.best_permutation = res_beta.opt.get('order')[0].astype(int)
        self.best_icor = res_beta.opt.get('icor')[0]
        self.best_error = res_beta.F[0]
        # print(f"Best permutation order: {permute_order}")
        # print(f"Beta {self.beta}, Best match icor: {best_match_icor}")

        # summarize the feature ids on each party
        party_cut_points = np.cumsum(self.evaluator.n_features_on_party)
        party_cut_points = np.insert(party_cut_points, 0, 0)
        self.best_feature_per_party = []
        for i in range(len(party_cut_points) - 1):
            start = party_cut_points[i]
            end = party_cut_points[i + 1]
            self.best_feature_per_party.append(np.sort(self.best_permutation[start:end]))
        assert (np.sort(np.concatenate(self.best_feature_per_party)) == np.arange(X.shape[1])).all()
        return self.best_feature_per_party

    def fit_split(self, X, **kwargs):
        self.fit(X, **kwargs)
        return self.split(X, **kwargs)

    def visualize(self, *args, **kwargs):
        return self.evaluator.visualize(*args, **kwargs)


class SimpleSplitter(Splitter):
    def __init__(self, num_parties):
        """
        Split a 2D dataset by equally dividing the features.
        :param num_parties: [int] number of parties
        """
        super().__init__(num_parties)

    def split_indices(self, n_features):
        """
        Split the indices of features into num_parties parts
        :param n_features: [int] number of features
        :return: [list of np.ndarray] indices of features for each party
        """
        indices = np.arange(n_features)
        return np.array_split(indices, self.num_parties)






import unittest

from collections import defaultdict
import numpy as np
import xgboost as xgb
from sklearn.datasets import make_classification

from src.vertibench.Evaluator import ImportanceEvaluator, CorrelationEvaluator


def generate_data():
    ###### Varying instances ######
    # Generate a small dataset with single instance
    X_1_10, y_1_10 = make_classification(n_samples=1, n_features=10, n_informative=5, n_redundant=3,
                                                   random_state=0, shuffle=True)
    X_2_10, y_2_10 = make_classification(n_samples=2, n_features=10, n_informative=5, n_redundant=3,
                                                    random_state=0, shuffle=True)

    # Generate a small dataset
    X_100_10, y_100_10 = make_classification(n_samples=100, n_features=10, n_informative=5, n_redundant=3,
                                                       random_state=0, shuffle=True)

    # Generate a large dataset
    X_10k_10, large_y_10k_10 = make_classification(n_samples=10000, n_features=10, n_informative=5,
                                                             n_redundant=3,
                                                             random_state=0, shuffle=True)

    ##### Varying features #####
    # Generate a dataset with single feature
    X_1k_2, y_1k_2 = make_classification(n_samples=1000, n_features=2, n_informative=2, n_redundant=0,
                                                   random_state=0, shuffle=True)

    # Generate a dataset with large number of features
    X_1k_1k, y_1k_1k = make_classification(n_samples=1000, n_features=1000, n_informative=500,
                                                     n_redundant=200,
                                                     random_state=0, shuffle=True)

    ###### Constant features ######
    # Generate a dataset with constant features
    X_1k_10_const, y_1k_10_const = make_classification(n_samples=1000, n_features=8, n_informative=5,
                                                                 n_redundant=1,
                                                                 random_state=0, shuffle=True)
    X_1k_10_const = np.concatenate((X_1k_10_const, np.zeros((1000, 2))), axis=1)

    # Generate a dataset with constant features with single feature
    X_1k_2_const, y_1k_2_const = np.zeros((1000, 2)), np.zeros((1000, 1))

    # parse local variables to create data
    data = {'1_10': (X_1_10, y_1_10),
            '2_10': (X_2_10, y_2_10),
            '100_10': (X_100_10, y_100_10),
            '10k_10': (X_10k_10, large_y_10k_10),
            '1k_2': (X_1k_2, y_1k_2),
            '1k_1k': (X_1k_1k, y_1k_1k),
            '1k_10_const': (X_1k_10_const, y_1k_10_const),
            '1k_2_const': (X_1k_2_const, y_1k_2_const)}

    return data

def split_data(data):
    splits = defaultdict(list)
    for k, (X, y) in data.items():
        n_features = X.shape[1]
        step = n_features // 5 if n_features > 5 else 1
        for i in range(1, n_features, step):
            splits[k].append(([X[:, :i], X[:, i:]], y))
    return splits


class TestImportanceEvaluator(unittest.TestCase):
    def setUp(self):
        self.evaluator = ImportanceEvaluator()
        self.small_sample_evaluator = ImportanceEvaluator(sample_rate=0.001)

        self.data = generate_data()
        self.splits = split_data(self.data)

    def test_evaluate(self):
        for k, split_list in self.splits.items():
            for Xs, y in split_list:
                with self.subTest(k=k):
                    model = xgb.XGBClassifier()
                    X = np.concatenate(Xs, axis=1)
                    model.fit(X, y)
                    scores, small_scores = None, None
                    if X.shape[0] * self.evaluator.sample_rate <= 1:
                        self.assertRaises(ValueError, self.evaluator.evaluate, Xs, model.predict)
                    else:
                        scores = self.evaluator.evaluate(Xs, model.predict)
                        self.assertEqual(len(scores), len(Xs))

                    if X.shape[0] * self.small_sample_evaluator.sample_rate <= 1:
                        self.assertRaises(ValueError, self.small_sample_evaluator.evaluate, Xs, model.predict)
                    else:
                        small_scores = self.small_sample_evaluator.evaluate(Xs, model.predict)
                        self.assertEqual(len(small_scores), len(Xs))
                    if scores is not None and small_scores is not None:
                        self.assertAlmostEqual(scores[0], small_scores[0], delta=0.1)

    def test_evaluate_feature(self):
        for k, (X, y) in self.data.items():
            with self.subTest(k=k):
                model = xgb.XGBClassifier()
                model.fit(X, y)
                scores, small_scores = None, None
                if X.shape[0] * self.evaluator.sample_rate <= 1:
                    self.assertRaises(ValueError, self.evaluator.evaluate_feature, X, model.predict)
                else:
                    scores = self.evaluator.evaluate_feature(X, model.predict)
                    self.assertEqual(len(scores), X.shape[1])

                if X.shape[0] * self.small_sample_evaluator.sample_rate <= 1:
                    self.assertRaises(ValueError, self.small_sample_evaluator.evaluate_feature, X, model.predict)
                else:
                    small_scores = self.small_sample_evaluator.evaluate_feature(X, model.predict)
                    self.assertEqual(len(small_scores), X.shape[1])

    def test_multi_party_evaluate(self):
        X, y = self.data['1k_1k']
        for n_parties in [2, 3, 4, 10, 100, 200, 500, 1000]:
            with self.subTest(n_parties=n_parties):
                Xs = np.array_split(X, n_parties, axis=1)
                model = xgb.XGBClassifier()
                model.fit(X, y)
                if X.shape[0] * self.evaluator.sample_rate <= 1:
                    self.assertRaises(ValueError, self.evaluator.evaluate, Xs, model.predict)
                else:
                    scores = self.evaluator.evaluate(Xs, model.predict)
                    self.assertEqual(len(scores), len(Xs))

                if X.shape[0] * self.small_sample_evaluator.sample_rate <= 1:
                    self.assertRaises(ValueError, self.small_sample_evaluator.evaluate, Xs, model.predict)
                else:
                    small_scores = self.small_sample_evaluator.evaluate(Xs, model.predict)
                    self.assertEqual(len(small_scores), len(Xs))


class TestCorrelationEvaluator(unittest.TestCase):
    def setUp(self):
        self.spearmanr_evaluator = CorrelationEvaluator(method='spearmanr', gpu_id=None)
        self.pearsonr_evaluator = CorrelationEvaluator(method='pearson', gpu_id=None)
        self.gpu_spearmanr_evaluator = CorrelationEvaluator(method='spearmanr', gpu_id=0)
        self.gpu_pearsonr_evaluator = CorrelationEvaluator(method='pearson', gpu_id=0)

        self.data = generate_data()
        self.splits = split_data(self.data)

    def subtest_fit_evaluate(self, Xs):
        if Xs[0].shape[0] <= 1:
            self.assertRaises(ValueError, self.spearmanr_evaluator.fit_evaluate, Xs)
            self.assertRaises(ValueError, self.pearsonr_evaluator.fit_evaluate, Xs)
            self.assertRaises(ValueError, self.gpu_spearmanr_evaluator.fit_evaluate, Xs)
            self.assertRaises(ValueError, self.gpu_pearsonr_evaluator.fit_evaluate, Xs)
        else:
            spearmanr_score = self.spearmanr_evaluator.fit_evaluate(Xs)
            pearsonr_score = self.pearsonr_evaluator.fit_evaluate(Xs)
            spearmanr_score_gpu = self.gpu_spearmanr_evaluator.fit_evaluate(Xs)
            pearsonr_score_gpu = self.gpu_pearsonr_evaluator.fit_evaluate(Xs)

            # score should be in range [-1, 1]
            self.assertGreaterEqual(spearmanr_score, -1)
            self.assertLessEqual(spearmanr_score, 1)
            self.assertGreaterEqual(pearsonr_score, -1)
            self.assertLessEqual(pearsonr_score, 1)
            self.assertGreaterEqual(spearmanr_score_gpu, -1)
            self.assertLessEqual(spearmanr_score_gpu, 1)
            self.assertGreaterEqual(pearsonr_score_gpu, -1)
            self.assertLessEqual(pearsonr_score_gpu, 1)

            # GPU score should be the same as CPU score
            self.assertAlmostEqual(spearmanr_score, spearmanr_score_gpu, delta=0.0001)
            self.assertAlmostEqual(pearsonr_score, pearsonr_score_gpu, delta=0.0001)

            # correlation matrix should be in range [-1, 1]
            self.assertTrue((self.spearmanr_evaluator.corr <= 1).all())
            self.assertTrue((self.spearmanr_evaluator.corr >= -1).all())
            self.assertTrue((self.pearsonr_evaluator.corr <= 1).all())
            self.assertTrue((self.pearsonr_evaluator.corr >= -1).all())
            self.assertTrue((self.gpu_spearmanr_evaluator.corr <= 1).all())
            self.assertTrue((self.gpu_spearmanr_evaluator.corr >= -1).all())
            self.assertTrue((self.gpu_pearsonr_evaluator.corr <= 1).all())
            self.assertTrue((self.gpu_pearsonr_evaluator.corr >= -1).all())

            # overall correlation matrix should be symmetric
            self.assertTrue(np.allclose(self.spearmanr_evaluator.corr, self.spearmanr_evaluator.corr.T))
            self.assertTrue(np.allclose(self.pearsonr_evaluator.corr, self.pearsonr_evaluator.corr.T))
            self.assertTrue(np.allclose(self.gpu_spearmanr_evaluator.corr.detach().cpu().numpy(),
                                        self.gpu_spearmanr_evaluator.corr.T.detach().cpu().numpy()))
            self.assertTrue(np.allclose(self.gpu_pearsonr_evaluator.corr.detach().cpu().numpy(),
                                        self.gpu_pearsonr_evaluator.corr.T.detach().cpu().numpy()))

    def test_fit_evaluate(self):
        for k, split_list in self.splits.items():
            for Xs, y in split_list:
                with self.subTest(k=k):
                    self.subtest_fit_evaluate(Xs)

    def test_fit_multi_evaluate(self):
        for k, split_list in self.splits.items():
            Xs = split_list[0][0]
            if Xs[0].shape[0] <= 1:
                self.assertRaises(ValueError, self.spearmanr_evaluator.fit, Xs)
                self.assertRaises(ValueError, self.pearsonr_evaluator.fit, Xs)
                self.assertRaises(ValueError, self.gpu_spearmanr_evaluator.fit, Xs)
                self.assertRaises(ValueError, self.gpu_pearsonr_evaluator.fit, Xs)
                continue
            else:
                self.spearmanr_evaluator.fit(Xs)
                self.pearsonr_evaluator.fit(Xs)
                self.gpu_spearmanr_evaluator.fit(Xs)
                self.gpu_pearsonr_evaluator.fit(Xs)

            for Xs, y in split_list:
                with self.subTest(k=k):
                    spearmanr_score_eval = self.spearmanr_evaluator.evaluate(Xs)
                    pearsonr_score_eval = self.pearsonr_evaluator.evaluate(Xs)
                    spearmanr_score_gpu_eval = self.gpu_spearmanr_evaluator.evaluate(Xs)
                    pearsonr_score_gpu_eval = self.gpu_pearsonr_evaluator.evaluate(Xs)

                    gpu_spearmanr_evaluator = CorrelationEvaluator(method='spearmanr', gpu_id=0)
                    spearmanr_score_gpu_fit_eval = gpu_spearmanr_evaluator.fit_evaluate(Xs)
                    gpu_pearsonr_evaluator = CorrelationEvaluator(method='pearson', gpu_id=0)
                    pearsonr_score_gpu_fit_eval = gpu_pearsonr_evaluator.fit_evaluate(Xs)

                    # eval and fit_eval should be the same
                    self.assertAlmostEqual(spearmanr_score_eval, spearmanr_score_gpu_fit_eval, delta=0.0001)
                    self.assertAlmostEqual(pearsonr_score_eval, pearsonr_score_gpu_fit_eval, delta=0.0001)
                    self.assertAlmostEqual(spearmanr_score_gpu_eval, spearmanr_score_gpu_fit_eval, delta=0.0001)
                    self.assertAlmostEqual(pearsonr_score_gpu_eval, pearsonr_score_gpu_fit_eval, delta=0.0001)

    def test_visualize(self):
        for k, split_list in self.splits.items():
            # before calling fit, visualize should raise error
            self.assertRaises(ValueError, self.spearmanr_evaluator.visualize)
            self.assertRaises(ValueError, self.pearsonr_evaluator.visualize)
            self.assertRaises(ValueError, self.gpu_spearmanr_evaluator.visualize)
            self.assertRaises(ValueError, self.gpu_pearsonr_evaluator.visualize)

        for k, split_list in self.splits.items():
            Xs = split_list[0][0]
            if Xs[0].shape[0] <= 1:
                self.assertRaises(ValueError, self.spearmanr_evaluator.fit, Xs)
                self.assertRaises(ValueError, self.pearsonr_evaluator.fit, Xs)
                self.assertRaises(ValueError, self.gpu_spearmanr_evaluator.fit, Xs)
                self.assertRaises(ValueError, self.gpu_pearsonr_evaluator.fit, Xs)
                continue

            for Xs, y in split_list:
                with self.subTest(k=k):
                    self.spearmanr_evaluator.fit(Xs)
                    self.pearsonr_evaluator.fit(Xs)
                    self.gpu_spearmanr_evaluator.fit(Xs)
                    self.gpu_pearsonr_evaluator.fit(Xs)

                    self.spearmanr_evaluator.visualize(save_path=f"test/tmp/{k}_spearmanr.png")
                    self.pearsonr_evaluator.visualize(save_path=f"test/tmp/{k}_pearsonr.png")
                    self.gpu_spearmanr_evaluator.visualize(save_path=f"test/tmp/{k}_gpu_spearmanr.png")
                    self.gpu_pearsonr_evaluator.visualize(save_path=f"test/tmp/{k}_gpu_pearsonr.png")

    def test_multi_party_fit_evaluate(self):
        X, y = self.data['1k_1k']
        for n_parties in [2, 3, 4, 10, 100, 200, 500, 1000]:
            with self.subTest(n_parties=n_parties):
                Xs = np.array_split(X, n_parties, axis=1)
                self.subtest_fit_evaluate(Xs)


if __name__ == '__main__':
    unittest.main()
